UNIT KomboBox;
{+H
---------------------------------------------------------------------------
  File        - KOMBOBOX.PAS
  Purpose     - Forces a combobox to show some Quicken-like behavior:
                1. the list box is visible when editing
                2. [ESCAPE] restores the original text
                3. the text value is autocompleted if it is a proper substring of list value.
                4. hides the style property (only allows the csDropDown style).
  Reference   - Expands on code provided by  Cresto Sylvain in his TCombo98
                component. Cresto Sylvain can be reached at:
                email: cresto@mygale.org
                http://www.mygale.org/~cresto/
                La Borde / 09230 / Contrazy / France
  Author      - Keith S. Brown
                E-Mail: keith_brown@juno.com
                2437 Bay Area Blvd #20, Houston, TX USA.
  Copyright    1998 Keith S. Brown
  Restrictions- This component is FREEWARE!
                No warranty is given by the author, expressed or implied.
  -*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-
  Revised     - 1998.0227 (KSB) Reformatted & renamed variables, etc. to standard format.
              - 1998.0302 (KSB) Modified to add Quicken-like behavior when selected.
              - 1998.0303 (KSB) Added proper behavior when <ENTER> is pressed.
              - 1998.0305 (KSB) Hide unwanted properties by inheriting from TCustomComboBox.
---------------------------------------------------------------------------}
INTERFACE
USES
{$IfDef Win32}
  Windows,
{$Else}
  WinTypes, WinProcs,
{$EndIf}
  Messages, SysUtils, Classes,  Graphics, Controls, Forms,    Dialogs,
  StdCtrls, DsgnIntf;

TYPE
  TOnComplete = PROCEDURE(aSender:TObject; aKey:WORD; VAR aCompleteWith:STRING; VAR aDoComplete:BOOLEAN) OF OBJECT;
  TKomboBoxAboutProperty= CLASS(TPropertyEditor)
  PRIVATE
    PROCEDURE   Edit; OVERRIDE;
    FUNCTION    GetAttributes: TPropertyAttributes; OVERRIDE;
    FUNCTION    GetValue: STRING; OVERRIDE;
  END {CLASS};

  TKomboBox = CLASS(TCustomComboBox)
  PRIVATE
    FfComplete: BOOLEAN;
    FsText    : STRING;
    FwKey     : WORD;
    FwMinLngth: WORD;
    { .. }
    FAbout    : TKomboBoxAboutProperty;
    FOnComplete:TOnComplete;
    PROCEDURE   SetComplete(aValue:BOOLEAN);
    PROCEDURE   SetMinLength(aValue:WORD);
  PUBLIC
    CONSTRUCTOR Create(aOwner : TComponent); OVERRIDE;
  PROTECTED
    PROCEDURE   Change; OVERRIDE;
    PROCEDURE   DoEnter; OVERRIDE;
    PROCEDURE   KeyDown(VAR aKey:WORD; aShift:TShiftState); OVERRIDE;
    PROPERTY    StoredText : STRING READ FsText WRITE FsText;
  PUBLISHED
    PROPERTY    About : TKomboBoxAboutProperty READ FAbout WRITE FAbout;
    PROPERTY    Color;
    PROPERTY    Complete : BOOLEAN READ FfComplete WRITE SetComplete;
    PROPERTY    Ctl3D;
    PROPERTY    Cursor;
    PROPERTY    DragCursor;
    PROPERTY    DropDownCount;
    PROPERTY    Enabled;
    PROPERTY    Font;
    PROPERTY    Height;
    PROPERTY    HelpContext;
    PROPERTY    Hint;
    PROPERTY    ItemHeight;
    PROPERTY    Items;
    PROPERTY    Left;
    PROPERTY    MaxLength;
    PROPERTY    MinLengthComplete : WORD READ FwMinLngth WRITE SetMinLength;
    PROPERTY    Name;
    PROPERTY    OnChange;
    PROPERTY    OnClick;
    PROPERTY    OnComplete : TOnComplete READ FOnComplete WRITE FOnComplete;
    PROPERTY    OnDblClick;
    PROPERTY    OnDragDrop;
    PROPERTY    OnDragOver;
    PROPERTY    OnDropDown;
    PROPERTY    OnEndDrag;
    PROPERTY    OnEnter;
    PROPERTY    OnExit;
    PROPERTY    OnKeyDown;
    PROPERTY    OnKeyPress;
    PROPERTY    OnKeyUp;
    PROPERTY    ParentColor;
    PROPERTY    ParentCtl3D;
    PROPERTY    ParentFont;
    PROPERTY    ParentShowHint;
    PROPERTY    PopupMenu;
    PROPERTY    ShowHint;
    PROPERTY    Sorted;
    PROPERTY    TabOrder;
    PROPERTY    TabStop;
    PROPERTY    Tag;
    PROPERTY    Text;
    PROPERTY    Top;
    PROPERTY    Visible;
    PROPERTY    Width;
  END {CLASS};

              {=================================================}
IMPLEMENTATION

{$IfDef Win32}
{$RESOURCE KOMBOBOX.RES}
{$Else}
{}FUNCTION  MakeLParam(awLow, awHigh: WORD): LongINT;
{+H
---------------------------------------------------------------------------
  Purpose     - 16 bit substitution for the MAKELPARAM macro.
                The MAKELPARAM macro creates an unsigned 32-bit value for
                use as an lParam parameter in a message. The macro
                concatenates two given 16-bit values.

  Parameters  -
        wLow    Specifies the low-order word of the new long value.
        wHigh   Specifies the high-order word of the new long value.

  Return Value- The return value is an unsigned 32-bit value.

  Remarks     - The MAKELPARAM macro is defined as follows:
                #define MAKELPARAM(l, h)   ((LPARAM) MAKELONG(l, h))
---------------------------------------------------------------------------}
  BEGIN
    Result := awLow OR awHigh SHL 16;
{}END {MakeLParam};
{$RESOURCE KOMBOB16.RES}
{$EndIf}

(*
procedure Register;
begin
  RegisterComponents('Standard', [TKomboBox]);
  RegisterPropertyEditor(TypeInfo(TKomboBoxAboutProperty), TKomboBox, 'ABOUT', TKomboBoxAboutProperty);
END;
*)

{ TKomboBox }

{}CONSTRUCTOR TKomboBox.Create(aOwner : Tcomponent);
  BEGIN
    INHERITED Create(aOwner);
    FfComplete := TRUE;
    FwKey := 0;
    FwMinLngth := 0;
    FsText := '';
{}END {Create};




{}PROCEDURE   TKomboBox.Change;
{+M
---------------------------------------------------------------------------
  Revised     - 1998.0227 (KSB) Reformatted code. Revised variable names.
              - 1998.0302 (KSB) Replaced the magic numbers 8 and 46 with the
                virtual key codes VK_BACK and VK_DELETE respectively. Revised
                code to reselect nearest match when deleting chars.
---------------------------------------------------------------------------}
  VAR
    lfD       : BOOLEAN;
    lfDoIt    : BOOLEAN;
    lw        : WORD;
    lwD       : WORD;
    ls        : STRING;
  BEGIN
    INHERITED Change;
    IF (FfComplete) AND ((Style=csDropDown) OR (Style=csSimple)) THEN BEGIN
      IF (Length(Text)>=FwMinLngth) AND (Items.Count > 0) THEN BEGIN
        {don't trigger on a "TAB" or "ESCAPE" key}
        IF (FwKey<>VK_TAB) AND (FwKey<>VK_ESCAPE) THEN BEGIN
          ls := UpperCase(Text);            {prepare for case insensitive search}
          lfD:= (FwKey=VK_BACK) OR (FwKey=VK_DELETE);

          FOR lw := 0 TO items.Count-1 DO BEGIN
            IF ls = UpperCase(Copy(items.strings[lw],1,Length(ls))) THEN BEGIN
              IF lfD THEN BEGIN             {"BACKSPACE" or "DELETE" key}
                ls := Text;
                ItemIndex := lw;
                Text := ls;
                SendMessage(Handle, CB_SETEDITSEL, 0, MakeLParam(Length(Text),Length(Text)));
              END ELSE BEGIN                {all keys, not [BACKSPACE,DELETE,TAB,ESCAPE]}
                lwD := Length(ls);
                ls := items.strings[lw];
                ItemIndex := lw;
                lfDoIt := TRUE;
                IF Assigned(FOnComplete) THEN BEGIN
                  FOnComplete(Self,FwKey,ls,lfDoIt);
                END {IF};

                IF lfDoIt THEN BEGIN
                  Text := ls;
                  SendMessage(Handle, CB_SETEDITSEL, 0, MakeLParam(lwD,Length(Text)));
                END {IF};
              END {IF};

              Break;
            END {IF};
          END {FOR};
        END {IF};
      END {IF};
    END {IF};
{}END {Change};




{}PROCEDURE   TKomboBox.DoEnter;
{+M
---------------------------------------------------------------------------
  Purpose     - Override the default DoEnter method.
  Remarks     - This method forces the following behaviors:
                1. the original text is saved and restored. Otherwise it is
                   cleared when the list box drops down.
                2. the list box will appear automatically when the control
                   receives the focus.
                3. It also resets the itemIndex to "unselected" (-1) [if
                   the control was visited before, it remembers the previous
                   selection from the list (default is 0) and forces the text
                   to contain that value unless the index is reset]
---------------------------------------------------------------------------}
  BEGIN
    StoredText := Text;
    INHERITED DoEnter;
    DroppedDown := TRUE;
    ItemIndex := -1;
    Text := StoredText;
{}END {DoEnter};




{}PROCEDURE   TKomboBox.KeyDown(VAR aKey: WORD; aShift: TShiftState);
{+M
---------------------------------------------------------------------------
  Purpose     - Override the default KeyDown method.
  Remarks     - This method forces the following behaviors:
                1. if the key is VK_TAB the list box is closed.
                2. if the key is VK_ESCAPE the list box is closed and the
                   original text value (the text value before the combo was
                   selected) is restored.
                3. if the key is VK_RETURN the closest match to the current
                   text value is selected if text is a proper substring of
                   any string in the list.
                4. For any other key, the list box is opened.
---------------------------------------------------------------------------}
  VAR
    ls        : STRING;
  BEGIN
    ls := Text;
    INHERITED KeyDown(aKey,aShift);
    FwKey := aKey;
    CASE FwKey OF
      VK_TAB    : DroppedDown := FALSE;
      VK_ESCAPE : BEGIN
        ItemIndex := -1;
        DroppedDown := FALSE;
        Text := StoredText;
        SendMessage(Handle, CB_SETEDITSEL, 0, MakeLParam(Length(Text),Length(Text)));
        Exit;
      END;

      VK_RETURN : BEGIN
        ls := Text;
        DroppedDown := FALSE;
        Text := ls;
        Change;
        Exit;
      END;

      ELSE BEGIN
        DroppedDown := TRUE;
      END;
    END {CASE};

    Text := ls;
{}END {KeyDown};




{}PROCEDURE   TKomboBox.SetComplete(aValue:BOOLEAN);
  BEGIN
    FfComplete := aValue;
{}END {SetComplete};




{}PROCEDURE   TKomboBox.SetMinLength(aValue:WORD);
  BEGIN
    FwMinLngth := aValue;
{}END {SetMinLength};



{ TKomboBoxAboutProperty }

{}PROCEDURE   TKomboBoxAboutProperty.Edit;
  BEGIN
    Application.MessageBox
      ('Derived from TComboBox98 By CRESTO Sylvain - 1997'#13
      +'E-Mail: cresto@mygale.org'#13
      +'WEB: http://www.mygale.org/~cresto/'#13
      +'TKomboBox by Keith Brown - 1998 - is FreeWare'#13
      +'Adds Quicken-like behavior to a TComboBox'#13
      +'E-Mail: keith_brown@juno.com'
      ,'TKomboBox version 1.0.0',MB_OK+ MB_ICONINFORMATION);
{}END {Edit};




{}FUNCTION    TKomboBoxAboutProperty.GetAttributes: TPropertyAttributes;
  BEGIN
    GetAttributes:=[paDialog, paReadOnly];
{}END {GetAttributes};




{}FUNCTION    TKomboBoxAboutProperty.GetValue: STRING;
  BEGIN
    GetValue:='(About)';
{}END {GetValue};


END {KomboBox}.
